#include <alsa/sound/type_compat.h>
#include <alsa/sound/uapi/emu10k1.h>
#ifndef __emu10k1_type_defined
#define __emu10k1_type_defined
typedef struct snd_emu10k1_fx8010_info emu10k1_fx8010_info_t;
typedef struct snd_emu10k1_fx8010_control_gpr emu10k1_fx8010_control_gpr_t;
typedef struct snd_emu10k1_fx8010_code emu10k1_fx8010_code_t;
typedef struct snd_emu10k1_fx8010_tram emu10k1_fx8010_tram_t;
typedef struct snd_emu10k1_fx8010_pcm_rec emu10k1_fx8010_pcm_t;
typedef struct emu10k1_ctl_elem_id emu10k1_ctl_elem_id_t;
#endif
